# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-08-13 13:07
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("djangocms_versioning", "0006_auto_20180809_1714"),
    ]

    operations = [
        migrations.AddField(
            model_name="version",
            name="created_by",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
            preserve_default=False,
        )
    ]
